#!/usr/bin/env python3

"""
Azure Blob Storage Sync Script

This script synchronizes local files with a specified Azure Blob Storage directory. It supports
file selection using masks or specific paths via --local-files and allows for exclusion patterns
using --exclude-files. Files can be compared based on size or checksum (MD5).

If a file with the same name already exists in Azure and differs from the local version,
the existing file is archived before uploading the new version.
"""

import argparse
import glob
import hashlib
import logging
import os
from typing import Dict, List, Tuple

from azure.storage.blob import BlobProperties, BlobServiceClient, ContainerClient
from azure.core.exceptions import ResourceExistsError
from colors import colors
from gitlab_logging import init_logging


def parse_arguments() -> argparse.Namespace:
    """
    Parse command-line arguments for syncing local files with Azure Blob Storage.

    """
    parser = argparse.ArgumentParser(description=__doc__)

    parser.add_argument(
        "--include-files",
        help='List of file paths or masks (e.g., "*.txt", /path/to/file1).',
        nargs="+",
        required=True,
    )
    parser.add_argument(
        "--exclude-files",
        help='List of file paths or masks to exclude (e.g., "*.log", temp*).',
        nargs="+",
        default=[],
    )
    parser.add_argument(
        "--azure-dir",
        help="Azure directory to sync with.",
        required=True,
    )
    parser.add_argument(
        "--azure-archive-dir",
        help="Azure directory for archiving old files.",
        required=True,
    )
    parser.add_argument(
        "--compare-by-size",
        help="Compare files by size.",
        action="store_true",
        default=True,
    )
    parser.add_argument(
        "--compare-by-hash",
        help="Compare files by hash (MD5).",
        action="store_true",
    )
    # Azure credentials
    parser.add_argument(
        "--azure-account-name",
        help="Azure Storage account name.",
        required=True,
    )
    parser.add_argument(
        "--azure-account-key",
        help="Azure Storage account key.",
        required=True,
    )
    parser.add_argument(
        "--azure-container-name",
        help="Azure Blob container name.",
        required=True,
    )
    # Verbosity
    parser.add_argument(
        "--verbose",
        help="Enable verbose logging.",
        action="store_true",
    )
    return parser.parse_args()


def get_blob_service_client(account_name: str, account_key: str) -> BlobServiceClient:
    """
    Initialize and return Azure BlobServiceClient.
    """
    connection_string = (
        f"DefaultEndpointsProtocol=https;AccountName={account_name};"
        f"AccountKey={account_key};EndpointSuffix=core.windows.net"
    )
    return BlobServiceClient.from_connection_string(connection_string)


def list_local_files(include_files: List[str], exclude_files: List[str]) -> Dict[str, str]:
    """
    List local files based on provided inclusion masks and exclusion patterns.
    Expands user directory (~) and handles both shell-expanded paths and wildcards.

    """
    included_files: List[str] = []
    excluded_files: List[str] = []

    logging.info("Listing local files...")

    # Expand ~ to full user directory if it is present in the path
    include_files = [os.path.expanduser(path) for path in include_files]
    exclude_files = [os.path.expanduser(path) for path in exclude_files]

    # Process inclusion files, handling both shell-expanded paths and raw wildcard patterns.
    # If the user doesn't quote wildcards (e.g., * or ?), Bash expands them into file lists before reaching the script.
    # If the wildcards are quoted, the script receives the pattern itself and needs to expand it using glob.
    # Example:
    #   Without quotes:  --local-files ~/data/*.txt   (Bash expands to /home/user/data/file1.txt, file2.txt)
    #   With quotes:     --local-files "~/data/*.txt" (Script receives "*.txt" and expands it internally)

    for file_or_pattern in include_files:
        if "*" in file_or_pattern or "?" in file_or_pattern:
            matched_files = glob.glob(file_or_pattern, recursive=True)
            included_files.extend(matched_files)
        elif os.path.exists(file_or_pattern):
            included_files.append(file_or_pattern)
        else:
            logging.warning(
                f"Warning: No matching file found for '{file_or_pattern}'. Please verify the file name or pattern."
            )

    # Process exclusion files (same logic as inclusion)
    for file_or_pattern in exclude_files:
        if "*" in file_or_pattern or "?" in file_or_pattern:
            matched_files = glob.glob(file_or_pattern, recursive=True)
            excluded_files.extend(matched_files)
        elif os.path.exists(file_or_pattern):
            excluded_files.append(file_or_pattern)

    included_files = sorted(included_files)
    excluded_files = sorted(excluded_files)

    logging.info(f"{len(included_files)} file(s) matched the inclusion pattern(s):")
    logging.info(
        "\t" + "\n\t".join(included_files),
        section_title="local_files",
        section_header="Local files",
    )

    if exclude_files:
        logging.info(f"{len(excluded_files)} file(s) matched the exclusion pattern(s):")
        logging.info(
            "\t" + "\n\t".join(excluded_files),
            section_title="excluded_files",
            section_header="Excluded files",
        )

    # Remove excluded files from included files
    final_files = set(included_files) - set(excluded_files)
    final_files = sorted(final_files)
    logging.info(f"{len(final_files)} file(s) selected after applying inclusion and exclusion patterns:")
    logging.info(
        "\t" + "\n\t".join(final_files),
        section_title="selected_files",
        section_header="Selected files",
    )

    return final_files


def list_azure_blobs(container_client: ContainerClient, azure_dir: str) -> Dict[str, BlobProperties]:
    """
    List all blobs in the specified Azure directory, excluding folder placeholders.

    """
    logging.info("Listing Azure blobs...")

    blob_list = {}
    blobs = container_client.list_blobs(name_starts_with=azure_dir)
    for blob in blobs:
        rel_path = os.path.relpath(blob.name, azure_dir)
        # Exclude subdirectories
        if "/" not in rel_path:
            blob_list[rel_path] = blob
    blob_list = dict(sorted(blob_list.items()))
    logging.info(f"{len(blob_list)} file(s) found in Azure directory: {azure_dir}")
    logging.info(
        "\t" + "\n\t".join(blob_list),
        section_title="blob_files",
        section_header="Blob files",
    )

    return blob_list


def calculate_md5(file_path: str) -> str:
    """
    Calculate the MD5 checksum of a file.
    """
    hash_md5 = hashlib.md5()
    with open(file_path, "rb") as f:
        for chunk in iter(lambda: f.read(4096), b""):
            hash_md5.update(chunk)
    return hash_md5.hexdigest()


def compare_files(
    local_files: List[str], azure_files: Dict[str, BlobProperties], compare_by_size: bool, compare_by_hash: bool
) -> Tuple[List[str], List[str]]:
    """
    Compare local files with Azure blobs to determine which need to be uploaded or archived.
    """
    logging.info("Comparing files...")

    files_to_upload = []
    files_to_archive = []

    for local_path in local_files:
        local_file_name = os.path.basename(local_path)
        local_size = os.path.getsize(local_path)
        local_md5 = calculate_md5(local_path) if compare_by_hash else None

        if local_file_name in azure_files:
            blob = azure_files[local_file_name]
            blob_size = blob.size
            blob_md5 = blob.content_settings.content_md5.hex() if blob.content_settings.content_md5 else None

            size_mismatch = compare_by_size and (local_size != blob_size)
            hash_mismatch = compare_by_hash and (local_md5 != blob_md5)

            if size_mismatch or hash_mismatch:
                files_to_archive.append(local_path)
                files_to_upload.append(local_path)
        else:
            files_to_upload.append(local_path)

    files_to_upload = sorted(files_to_upload)
    files_to_archive = sorted(files_to_archive)
    logging.info(f"{len(files_to_upload)} file(s) need to be uploaded.")
    logging.info(
        "\t" + "\n\t".join(files_to_upload),
        section_title="upload_files",
        section_header="Upload files",
    )
    if files_to_archive:
        logging.info(f"{len(files_to_archive)} file(s) will be archived before uploading new versions.")
        logging.info(
            "\t" + "\n\t".join(files_to_archive),
            section_title="archive_files",
            section_header="Archive files",
        )

    return files_to_upload, files_to_archive


def upload_and_archive(
    container_client: ContainerClient,
    azure_dir: str,
    archive_dir: str,
    files_to_upload: List[str],
    files_to_archive: List[str],
) -> None:
    """
    Archive outdated blobs and upload new or updated files to Azure Blob Storage.
    """

    for local_path in files_to_upload:
        local_file_name = os.path.basename(local_path)
        target_blob = f"{azure_dir}/{local_file_name}"

        if local_path in files_to_archive:
            local_file_name = os.path.basename(local_path)
            source_blob = f"{azure_dir}/{local_file_name}"
            archive_blob = f"{archive_dir}/{local_file_name}"

            logging.info(f"Archiving existing blob from {source_blob} to {archive_blob}")
            source_blob_client = container_client.get_blob_client(blob=source_blob)
            archive_blob_client = container_client.get_blob_client(blob=archive_blob)

            # Copy blob to archive
            archive_blob_client.start_copy_from_url(source_blob_client.url)
            # Delete original blob
            source_blob_client.delete_blob()

        logging.info(f"Uploading {local_path} to Azure blob: {target_blob}")
        blob_client = container_client.get_blob_client(blob=target_blob)
        with open(local_path, "rb") as data:
            try:
                blob_client.upload_blob(data, overwrite=False)

            except ResourceExistsError:
                error_message = (f"Upload failed: The file '{blob_client.blob_name}' already exists in Azure Blob Storage. "
                                f"Expected it to be archived before upload. Please check the archival process.")
                logging.error(error_message)
                raise RuntimeError(error_message)

def main() -> None:
    """
    Main entry point.
    """
    args = parse_arguments()
    init_logging(args.verbose, message_color=colors.fg.yellow)
    # Suppress INFO logs from Azure, only show warnings and errors
    logging.getLogger("azure").setLevel(logging.WARNING)

    # Log modified arguments to prevent exposing the Azure account key in the logs
    logging.add_masked_value(args.azure_account_key)
    logging.debug(args, section_title="args", section_header="Arguments")

    blob_service_client = get_blob_service_client(args.azure_account_name, args.azure_account_key)
    container_client = blob_service_client.get_container_client(args.azure_container_name)

    # List local files
    local_files = list_local_files(args.include_files, args.exclude_files)

    # List blob files
    azure_files = list_azure_blobs(container_client, args.azure_dir)

    # Comparing files
    files_to_upload, files_to_archive = compare_files(
        local_files, azure_files, args.compare_by_size, args.compare_by_hash
    )

    if not files_to_upload and not files_to_archive:
        logging.info("All files are up to date. No synchronization needed.")
        return

    # Perform the synchronization after checks
    logging.start_section("uploading", "Uploading..", collapsed=False)
    upload_and_archive(container_client, args.azure_dir, args.azure_archive_dir, files_to_upload, files_to_archive)
    logging.end_section("uploading", "Uploading..")

    logging.info("File synchronization completed successfully.")


if __name__ == "__main__":
    main()
