#!/usr/bin/env python3
""" 

Simple changelog generator for SECO gitlab projects.

Queries merge request from gitlab and outputs as sorted list in
markdown format.

Releases are tags on the given branch in the manifest project.
Changes to list are merged mergerequests with the given branch
as target. Match with releases is done by the timestamp (merged_at)
in comparison with the tags timestamp.

"""


import argparse
import datetime
import logging
import sys

import gitlab as gl

import common

__author__ = "Jonas Höppner"
__email__ = "jonas.hoeppner@garz-fricke.com"

GITLAB_SERVER = "https://git.seco.com"

GITLAB_TIMEFORMAT = "%Y-%m-%dT%H:%M:%S.%f%z"
TIMEFORMAT = "%Y-%m-%d %H:%M"

verbose = 0


def decode_timestamp(t):
    timestamp = datetime.datetime.strptime(t, GITLAB_TIMEFORMAT)
    return timestamp

def get_project_branch_from_integration(gitlab, project_path, manifest_branch):

    # Get project ID from project path
    project_id = gitlab.projects.get(project_path).id

    # Get CICD variables for the project
    variables = gitlab.projects.get(project_id).variables.list()

    # Initialize dictionary to store project branches corresponding to manifest branches
    branch_project_dict = {}

    # Parse INTEGRATION variable and extract 
    for variable in variables:
        if variable.key == 'INTEGRATION':
            integration_value = variable.value.split('\n')
            for line in integration_value:
                parts = line.split(':')
                branch_project, manifest, branch_manifest = parts
                if branch_manifest == manifest_branch:
                    if branch_manifest not in branch_project_dict:
                        branch_project_dict[branch_manifest] = []
                    branch_project_dict[branch_manifest].append(branch_project)
    
    # If INTEGRATION variable is not defined or no matching branch_project found, use manifest_branch as branch_project
    if not branch_project_dict:
        branch_project_dict = {manifest_branch: [manifest_branch]}

    # Return branch_projects corresponding to branch_manifest
    return branch_project_dict.get(manifest_branch, [])

class Project:
    def __init__(self, project):
        self.project = project

    def __str__(self):
        return "## Project " + self.project.name + "\n"

    def withlink(self):
        return (
            "\n\n## Project [" + self.project.name + "](" + self.project.web_url + ")\n"
        )

    def __eq__(self, p):
        if not p:
            return False
        return self.project.id == p.project.id


class Tag:
    def __init__(self, tag):
        self.name = tag.name
        self.message = tag.message
        self.commit = tag.commit
        """
        The tags timestamp is a little more complicated it normally points 
        to the tagged commit's timestamps. But the merge happens later.
        To handle this, the relelated mergerequest is found by comparing the
        sha's and also take the merged_at timestamp.
        """
        self.timestamp = decode_timestamp(tag.commit["created_at"])

        """
        The mr which introduced the taged commit
        as gitlab-python does not support the V5 API yet
        this is added later when traversing the mrs anyway
        with V5 Api: https://docs.gitlab.com/ee/api/commits.html#list-merge-requests-associated-with-a-commit
        """
        self.mergerequest = None
        logging.debug(self.name + "  -- " + self.commit["id"])

    def __str__(self):
        return self.name + " " + self.timestamp.strftime(TIMEFORMAT)

    def add_mergerequest(self, m):
        if self.mergerequest:
            return
        if m.mr.sha == self.commit["id"]:
            self.mergerequest = m
            # Update timestamp
            # The tag points to the commit, but the merge of the merge request may has happend later
            # as the commit, so the merged_at date is relevant. Otherwise the tagged commit and may be
            # more end up in the wrong release
            new_timestamp = decode_timestamp(self.mergerequest.mr.merged_at)
            logging.debug("Found matching merge request for %s", self)
            logging.debug(" - %s", self.timestamp.strftime(TIMEFORMAT))
            logging.debug(" - %s", new_timestamp.strftime(TIMEFORMAT))
            self.timestamp = new_timestamp

    def header(self):
        return (
            "\n\n\n# Release "
            + self.name
            + "\n\nreleased at "
            + self.timestamp.strftime(TIMEFORMAT)
            + "\n\n"
        )


class DummyTag:
    def __init__(
        self, name, message, date=datetime.datetime.now(tz=datetime.timezone.utc)
    ):
        self.name = name
        self.message = message
        self.timestamp = date

    def header(self):
        return "\n\n\n# " + self.name + "\n\n"

    def add_mergerequest(self, m):
        # Needed as interface but does nothing
        pass


class Release:
    """Store some release data"""

    def __init__(self, tag):
        self.tag = tag
        self.mergerequests = []

    def add_mergerequest(self, m):
        # Check if this merge_request is related to the tag
        self.tag.add_mergerequest(m)

        # Adds a mergerequest to the project, but uses some filtering
        # Ignore automated merge requests except for those regarding the 
        # BSP since they can be used to access the original merge request in 
        # the BSP repository itself
        bsp_repo = ["u-boot-seco-imx", "linux-seco-imx", "imx-mkimage", "u-boot-seco-rk", "linux-seco-rk", "u-boot-seco-mtk", "linux-seco-mtk"]
        if m.mr.author["username"] == "gitbot":
            if not any(el in m.mr.title for el in bsp_repo):
                return False
        
        # Exclude MRs containing CICD reference in the title
        if "CICD" in m.mr.title:
            return False
        if "CI/CD" in m.mr.title:
            return False
        # With the movement to git.seco.com the MRs owned by
        # the guf-gitbot have been transfered to tobias
        # As it is not possible to change the owner back
        # to gitbot we need an extra filter here on the
        # branch name
        if m.mr.source_branch.startswith("integrate/"):
            return False

        # Timestamp is not in this release
        if self.tag.timestamp < m.timestamp:
            return False

        # Remove duplicates, don't print the same title
        # twice in the same project and release
        if any(
            a.mr.title == m.mr.title and a.project == m.project
            for a in self.mergerequests
        ):
            return True

        self.mergerequests.append(m)
        return True

    def header(self):
        return self.tag.header()

    def description(self):
        m = self.tag.message
        if not m:
            return ""
        return m

    def __str__(self):
        return self.tag.name


class MergeRequest:
    def __init__(self, mr, p):
        self.mr = mr
        self.project = p
        self.timestamp = decode_timestamp(self.mr.merged_at)
        logging.debug("\nMergeRequest:")
        logging.debug(mr)

    def __str__(self):
        return self.mr.title

    def withlink(self):
        out = self.mr.title + " [" + self.mr.reference + "](" + self.mr.web_url + ")"
        return out


def main(args):

    parser = argparse.ArgumentParser(description=__doc__, usage="%(prog)s [OPTIONS]")

    parser.add_argument(
        "--gitlab-url",
        help="""URL to the GitLab instance""",
        dest="gitlab_url",
        default=common.GITLAB_URL,
    )
    parser.add_argument(
        "--token",
        help="""GitLab REST API private access token""",
        dest="token",
        required=True,
    )
    parser.add_argument(
        "--project",
        help="""Project to get merge requests from
                (can be passed multiple times, tags are read from the first one)""",
        dest="project",
        action="append",
        required=True,
    )
    parser.add_argument(
        "-b",
        "--branch",
        help="""Branch to work on""",
        dest="branch",
        required=True,
    )

    parser.add_argument(
        "-v",
        "--verbose",
        action="count",
        dest="verbose",
        default=0,
        help=("Increase verbosity."),
    )

    options = parser.parse_args(args)
    if options.verbose:
        logging.basicConfig(level=logging.DEBUG)

    logging.debug(options)
    gitlab = gl.Gitlab(options.gitlab_url, private_token=options.token)

    projects = []
    for project in options.project:
        projects.append(Project(gitlab.projects.get(project)))

    releases = []
    for t in projects[0].project.tags.list(): #search=options.branch
        releases.append(Release(Tag(t)))

    # Add dummy release with date today for new untaged commits
    releases.append(
        Release(
            DummyTag(
                "Not yet released",
                "Merge Request already merged into "
                + options.branch
                + " but not yet released.",
            )
        )
    )

    # Sort by date, oldest first
    releases = sorted(releases, key=lambda d: d.tag.timestamp, reverse=False)

    for p in projects:
        project_namespace=p.project.path_with_namespace
        branch_list=get_project_branch_from_integration(gitlab, project_namespace, options.branch)
        for branch in branch_list:
            #print(f"For this project {project_namespace} we search for the MRs on this branch {branch}")
            for mr in p.project.mergerequests.list(
                scope="all", state="merged", target_branch=branch, per_page="10000"
            ):
                m = MergeRequest(mr, p)
                for r in releases:
                    if r.add_mergerequest(m):
                        break
    
            #Iterate again over MR because of the kirkstone/develop -> kirkstone changement
            old_branch="kirkstone/develop"
            for mr in p.project.mergerequests.list(
                scope="all", state="merged", target_branch=old_branch, per_page="10000"
            ):
                m = MergeRequest(mr, p)
                for r in releases:
                    if r.add_mergerequest(m):
                        break

    # Sort by date, newest first
    releases = sorted(releases, key=lambda d: d.tag.timestamp, reverse=True)

    for r in releases:
        # Don't show empty releases/tags
        if not len(r.mergerequests):
            continue

        print(r.header())
        print(r.description())

        current_project = None
        for m in r.mergerequests:
            if m.project != current_project:
                current_project = m.project
                print(current_project.withlink())
            print(" - ", m.withlink())


if __name__ == "__main__":
    main(sys.argv[1:])
