#!/usr/bin/env python3

import argparse
import copy
import logging
import sys
import tempfile

import common
from colors import colors
from download_job_artifacts import download_job_artifact
from get_parent_pipeline import get_parent_pipeline
from get_pipeline_jobs import get_pipeline_jobs
from gitlab import GitlabGetError
from gitlab.client import Gitlab
from gitlab.v4.objects import Project, ProjectPipelineJob, ProjectPipelineVariableManager
from gitlab_logging import init_logging


def parse_arguments() -> argparse.Namespace:
    """
    Parse command-line arguments.

    """
    parser = argparse.ArgumentParser(description=__doc__, usage="%(prog)s [OPTIONS]")

    parser.add_argument(
        "--gitlab-url",
        help="""URL to the GitLab instance""",
        dest="gitlab_url",
        action="store",
        default=common.GITLAB_URL,
    )
    parser.add_argument(
        "--token",
        help="""GitLab REST API private access token""",
        dest="token",
        required=True,
    )
    parser.add_argument(
        "--project",
        action="store",
        dest="project",
        help="Specify the project by either by id or by path.",
        required=True,
    )
    parser.add_argument(
        "--pipeline",
        action="store",
        dest="pipeline",
        help="Specify the pipeline by id.",
        required=True,
    )
    parser.add_argument(
        "--release-pipeline-name",
        action="store",
        dest="release_pipeline_name",
        help="Specifies the name of the release pipeline to be excluded from the release data collection.",
        required=True,
    )
    parser.add_argument(
        "--deploy-stage",
        action="store",
        dest="deploy_stage",
        help="Specify the deploy stage to load artifacts from.",
        required=True,
    )
    parser.add_argument(
        "--export-to",
        action="store",
        dest="export_to",
        help="Export collected data to file.",
        required=True,
    )
    # Verbosity
    parser.add_argument(
        "--verbose",
        help="Enable verbose logging.",
        action="store_true",
    )
    return parser.parse_args()


def get_job_env_variables(gitlab: Gitlab, project: Project, job: ProjectPipelineJob, env_file: str):
    """Get a list of variables from the env_file of the given job.

    Args:
        gitlab: Gitlab object
        project: Gitlab project
        job: GitLab job
        env_file: name of the GitLab dotenv file to read variables from

    Returns:
        Dictionary of variables with 'name' and an set of 'value' fields
    """
    variables = {}

    with tempfile.NamedTemporaryFile() as target_file:
        logging.debug(f"Downloading file {env_file} from job {job.name}")
        try:
            download_job_artifact(
                gitlab,
                target_file.name,
                env_file,
                job.id,
                project=project,
            )
        except GitlabGetError:
            print(f"Artifact '{env_file}' not found in {job.web_url}", file=sys.stderr)

        # Additionally, filter out lines that do not resemble
        # environment variables (i.e., lines that do not contain '=')
        lines = (line.decode("utf-8").rstrip() for line in target_file if "=" in line.decode("utf-8"))

        for line in lines:
            name, value = line.partition("=")[::2]
            value = value.strip("'\" ")
            if name not in variables.keys():
                variables[name] = set()
            for elem in value.split(" "):
                if len(elem) > 0:
                    variables[name].add(elem)

    return variables


def add_suffix_to_dict_keys(dictionary: dict, suffix: str):
    """Add a suffix to every key of a dictionary.

    Args:
        dictionary: the dictionary
        suffix: the suffix to add

    Returns:
        Dictionary with suffixed keys
    """
    return {f"{key}{suffix}": dictionary[key] for key in dictionary}


def get_pipeline_variable_value(variables: ProjectPipelineVariableManager, key: str):
    """Lookup a ProjectPipelineVariable in a list by its key and return its value

    Args:
        variables: list of variables
        key: key of the variable to lookup

    Returns:
        Value
    """
    try:
        variable = next((variable for variable in variables.list(as_list=False) if variable.key == key))
    except StopIteration:
        return None

    return variable.value


def merge_variables(first: dict[str, set[str]], second: dict[str, set[str]]) -> dict[str, set[str]]:
    """Merge two dictionaries of variables, each variable stored as set.
       If a variable exists in both dictionaries the values are merged into one
       without duplicates.

    Args:
        first: a dictionary containing variables
        second: a dictionary containing variables

    Returns:
        A dictionary containing merged variables
    """
    merged = copy.deepcopy(first)
    for key, value in second.items():
        if key not in first:
            merged[key] = value
        else:
            merged[key] |= value
    return merged


def main() -> None:
    """
    Main entry point.
    """
    args = parse_arguments()
    init_logging(args.verbose, message_color=colors.fg.yellow)

    # Suppress DEBUG logs from urllib3
    logging.getLogger("urllib3.connectionpool").setLevel(logging.WARNING)

    # Mask sensitive arguments
    logging.add_masked_value(args.token)
    logging.debug(args, section_title="args", section_header="Arguments")

    gitlab = Gitlab(args.gitlab_url, private_token=args.token)
    project = common.get_project(gitlab, args.project)
    pipeline = project.pipelines.get(args.pipeline, retry_transient_errors=True)

    pipelines = []
    parent_pipeline = get_parent_pipeline(project, pipeline)
    if parent_pipeline:
        # Get all parallel child pipelines exept release pipeline
        logging.start_section("retrieve_child_pipelines", "Retrieving child pipelines..", collapsed=False)
        for bridge in parent_pipeline.bridges.list(as_list=False):
            logging.debug(f"Check downstream: '{bridge.name}' ({bridge.web_url})")
            if bridge.name == args.release_pipeline_name:
                logging.debug("Skip release downstream.")
                continue
            child_pipeline = project.pipelines.get(bridge.downstream_pipeline["id"], retry_transient_errors=True)
            logging.debug(f"Collect child pipeline: {child_pipeline.id}")
            pipelines.append({"pipeline": child_pipeline, "name": bridge.name})
        logging.end_section("retrieve_child_pipelines", "Retrieving child pipelines..")

    # Loop over all found pipelines and combine build and deploy variables from them
    logging.start_section("retrieve_variables", "Retrieving deploy jobs data..", collapsed=False)
    variables = {}
    for pipeline in pipelines:

        pipeline_variables = {}
        logging.debug(f"Getting jobs from pipeline '{pipeline['name']}' ({pipeline['pipeline'].web_url})")

        # Get all successful jobs from deploy stage
        logging.debug(f"Collecting from stage: '{args.deploy_stage}'")
        deploy_jobs = get_pipeline_jobs(gitlab, project, pipeline["pipeline"].id, stage=args.deploy_stage)
        logging.debug(f"{len(deploy_jobs)} jobs collected")
        successful_deploy_jobs = [job for job in deploy_jobs if job.status == "success"]
        if not successful_deploy_jobs:
            logging.debug(f"No successful jobs found in stage '{args.deploy_stage}'")
            continue

        logging.debug(f"{len(successful_deploy_jobs)} successful jobs collected")

        logging.debug("Dovnloading artifacts..")
        for deploy_job in successful_deploy_jobs:
            # Get job suffix after first "-" of the job name
            job_suffix = f"{deploy_job.name[deploy_job.name.find('-'):]}"

            job_env_variables = get_job_env_variables(gitlab, project, deploy_job, "deploy.env")

            # Append job suffix to the variables
            pipeline_variables |= add_suffix_to_dict_keys(job_env_variables, job_suffix.replace("-", "_"))

            # Furthermore merge all machine-specific values into one space-separated variable.
            pipeline_variables = merge_variables(pipeline_variables, job_env_variables)

        variables = merge_variables(variables, pipeline_variables)

    logging.end_section("retrieve_variables", "Retrieving build and deploy data..")
    logging.start_section("export_data", f"Export data to file {args.export_to}..")
    # Write all variables in a shell compatible form
    with open(args.export_to, "w", encoding="utf-8") as out_file:
        for k, v in sorted(variables.items()):
            # for k, v in variables.items():
            data = f'{k}="{" ".join(v)}"'
            out_file.write(data + "\n")
            logging.info(data)
    logging.end_section("export_data", f"Export data to file {args.export_to}..")


if __name__ == "__main__":
    main()
