#!/usr/bin/env python3
"""
    Creates release page on GitLab or appends assets to an existing release page.

"""

import common
import argparse
import logging
import glob
import re

from gitlab import Gitlab, GitlabGetError, GitlabCreateError

__author__ = "Andrii Sosiuk"
__email__ = "andrii.sosiuk@seco.com"


def parse_asset_files(files):
    """
    Parses asset files and returns a list of assets
    formatted according to GitLab's requirements.
    """
    logging.debug("Start parsing asset files: %s", files)
    files_data = []
    assets_list = []

    for file in files:
        logging.debug("Reading file:%s", file)
        with open(file, "r", encoding="utf-8") as fl:
            file_data = [file_line.strip() for file_line in fl.readlines()]
            logging.debug("Read %d lines", len(file_data))
            files_data.extend(file_data)
    logging.debug("Extracted %d lines from all files.", len(files_data))

    logging.debug("Parsing lines..")
    for line in files_data:
        if "https" in line:
            logging.debug("Found https:// record in line:")
            logging.debug(line)
            asset_file_name = re.sub(r"^.*/", "", line).split("?")[0]
            logging.debug("Creating asset: %s", {"name": asset_file_name, "url": line})
            assets_list.append({"name": asset_file_name, "url": line})

    logging.debug("%d assets were created.", len(assets_list))

    return assets_list

verbose = 0

def main():
    parser = argparse.ArgumentParser()

    parser.add_argument(
        "--gitlab-url",
        help="""URL to the GitLab instance""",
        dest="gitlab_url",
        action="store",
        default=common.GITLAB_URL,
    )
    parser.add_argument(
        "--token",
        help="""GitLab REST API private access token""",
        dest="token",
        required=True,
    )
    parser.add_argument(
        "--project",
        action="store",
        dest="project",
        help="Specify the project either by id or by path.",
        required=True,
    )
    parser.add_argument(
        "--name",
        action="store",
        dest="release_name",
        help="Specify the name for the release.",
        required=True,
    )
    parser.add_argument(
        "--tag-name",
        action="store",
        dest="tag_name",
        help="Specify the TAG name for the release.",
        required=True,
    )
    parser.add_argument(
        "--description-file",
        action="store",
        dest="release_description_file",
        help="Specify the file with release notes.",
        required=False,
        default="",
    )
    parser.add_argument(
        "--asset-files",
        action="store",
        dest="asset_files",
        nargs="+",
        help=(
            "Files containing lists of assets. "
            "Supports multiple file paths and wildcards (e.g., '*.txt')."
        ),
        required=False,
        default=[],
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        help="""Increase verbosity.""",
    )

    args, _ = parser.parse_known_args()
    if args.verbose:
        logging.basicConfig(
            level=logging.DEBUG,
        )

    logging.debug(args)
    gitlab = Gitlab(args.gitlab_url, private_token=args.token)
    project = common.get_project(gitlab, args.project)

    # Parse assets files
    asset_files_list = []
    for pattern in args.asset_files:
        asset_files_list.extend(glob.glob(pattern))

    if not asset_files_list:
        print("Warning: No asset files found!")
        print("Please check the command parameters.")
    assets = parse_asset_files(asset_files_list)

    # Read release description
    release_description = ""
    if args.release_description_file:
        with open(args.release_description_file, "r", encoding="utf-8") as description_file:
            release_description = description_file.read()

    # Check if release already exists

    try:
        release = project.releases.get(args.tag_name)
        print(f"Release {args.tag_name} already exists. Updating assets...")
    except GitlabGetError:
        logging.debug("No release found with tag: %s", args.tag_name)
        print(f"Creating a new release: {args.tag_name}...")
        release = project.releases.create(
            {
                "name": f"{args.release_name}",
                "tag_name": args.tag_name,
                "description": release_description,
            }
        )

    # Add assets to the release
    new_assets = 0
    for asset in assets:
        try:
            logging.debug("Adding asset: %s", asset["name"])
            release.links.create(asset)
            print(f"Added asset: {asset['name']}")
            new_assets += 1

        except GitlabCreateError as error:
            print(f"Failed to add asset: {asset['name']}")
            print(error)

    print(f"Release {args.tag_name} completed.")
    if new_assets:
        print(f"{new_assets} new assets have been added.")
    else:
        print("No new assets were added.")

    print(f"Release page: {release._links['self']}")

if __name__ == "__main__":
    main()
