#!/usr/bin/env python3
import argparse
import re
import sys
from gitlab import Gitlab, GitlabGetError
from gitlab.v4.objects import Group


def get_integration_sources(manifest_project: str, manifest_branch: str, group: Group):
    """
    Get a list of projects and branches in the given group which are configured for
    automatic integration into the given branch of the given manifest project.
    """
    integration_sources = []
    gitlab = group.manager.gitlab

    # Recurse into subgroups
    for g in group.subgroups.list():
        subgroup = gitlab.groups.get(g.id)
        integration_sources += get_integration_sources(
            manifest_project, manifest_branch, subgroup
        )

    # Regex to check INTEGRATION variable against (include also layers integration)
    architecture = ["imx", "rk", "mtk"]
    regex = [f":{manifest_project}:{manifest_branch}$"]
    manifest_subgroup = '/'.join(manifest_project.split('/')[:-1])
    for arch in architecture:
        regex.append(f"{manifest_subgroup}/layers/seco/meta-seco-{arch}:{manifest_branch}$")

    for project in group.projects.list():
        try:
            project = gitlab.projects.get(project.id)
            if not project.archived and project.jobs_enabled:
                integrations = project.variables.get("INTEGRATION").value
                for integration in integrations.splitlines():
                    for reg in regex:
                        if re.search(reg, integration):
                            source_branch = integration.split(":")[0]
                            integration_sources.append(
                                {
                                    "project": project.path_with_namespace,
                                    "branch": source_branch,
                                }
                            )
        except GitlabGetError as e:
            if e.response_code == 404:  # not found
                pass
            elif e.response_code == 403:  # forbidden
                sys.exit(
                    (
                        "ERROR: could not get INTEGRATION variable of project %s\n"
                        % project.path_with_namespace
                    )
                    + e.error_message
                )
            else:
                raise

    return integration_sources


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--gitlab-url",
        help="""URL to the GitLab instance""",
        dest="gitlab_url",
        required=True,
    )
    parser.add_argument(
        "--token",
        help="""GitLab REST API private access token""",
        dest="token",
        required=True,
    )
    parser.add_argument(
        "--manifest-project",
        help="""name of the manifest project""",
        dest="manifest_project",
        required=True,
    )
    parser.add_argument(
        "--manifest-branch",
        help="""manifest branch""",
        dest="manifest_branch",
        required=True,
    )
    parser.add_argument(
        "--group",
        help="""group path or id to limit search scope to""",
        dest="group",
        required=True,
    )

    args, _ = parser.parse_known_args()

    gitlab = Gitlab(args.gitlab_url, private_token=args.token)
    group = gitlab.groups.get(args.group)

    integration_sources = get_integration_sources(
        args.manifest_project,
        args.manifest_branch,
        group,
    )

    for source in integration_sources:
        print("%s:%s" % (source["project"], source["branch"]))


if __name__ == "__main__":
    main()
