#!/usr/bin/env python3
import argparse
from typing import Union

from gitlab.client import Gitlab
from gitlab.v4.objects import Project, ProjectPipeline

import common


def get_parent_pipeline(
    project: Project, pipeline: ProjectPipeline
) -> Union[ProjectPipeline, None]:
    """Get the parent pipeline within the same project

    Args:
        pipeline: Pipeline to get the parent of

    Returns:
        Parent pipeline if found, None else
    """

    # Get all pipelines for the same ref
    pipelines = project.pipelines.list(ref=pipeline.ref, as_list=False)

    # Browse through all pipelines and check if given pipeline is a child pipeline of it
    for p in pipelines:
        for bridge in p.bridges.list(as_list=False):
            if bridge.downstream_pipeline["id"] == pipeline.id:
                return p

    # Else return nothing
    return None


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--gitlab-url",
        help="""URL to the GitLab instance""",
        dest="gitlab_url",
        default=common.GITLAB_URL,
    )
    parser.add_argument(
        "--token",
        help="""GitLab REST API private access token""",
        dest="token",
        required=True,
    )
    parser.add_argument(
        "--project",
        help="""name of the GitLab project""",
        dest="project",
        required=True,
    )
    parser.add_argument(
        "--pipeline",
        help="""Pipeline to get parent pipeline of""",
        dest="pipeline",
        required=True,
    )

    args, _ = parser.parse_known_args()

    gitlab = Gitlab(args.gitlab_url, private_token=args.token)
    project = gitlab.projects.get(args.project)
    pipeline = project.pipelines.get(args.pipeline)

    parent_pipeline = get_parent_pipeline(project, pipeline)

    if not parent_pipeline:
        exit("No parent pipeline found")

    print(parent_pipeline.web_url)


if __name__ == "__main__":
    main()
