#!/usr/bin/env python3

"""
Enhanced logging capabilities for scripts running in GitLab CI/CD pipelines.

Features:
- Configures logging with different verbosity levels.
- Redirects logging output to stdout to maintain proper log order and
  ensures log messages appear in real-time by adjusting buffering settings.
- Supports GitLab CI collapsible sections using `section_start` and `section_end`.
- Allows color customization with predefined colors from `colors.py`.
- For non-CI environments, logs are still split into sections using delimiters.
- Provides automatic masking of sensitive data in log messages.

The masking feature allows users to define sensitive values that will be replaced
with a customizable masking string whenever they appear in log messages.

Usage:
    import gitlab_logging

    # Initialize logging for GitLab CI/CD
    gitlab_logging.init_logging(verbose=False)

    # Start a collapsible section in GitLab logs
    logging.start_section(title="install_dependencies", header="Installing Dependencies", header_color=colors.fg.cyan)
    logging.info("Fetching packages...")
    logging.end_section(title="install_dependencies", header="Dependencies Installed", header_color=colors.fg.cyan)

    # Mask sensitive data
    my_pass = "MySensitiveValue"
    logging.add_masked_value(my_pass)
    logging.info(f"This is my password: {my_pass}")

To switch between CI and plain mode, set the "CI" environment variable to "true" or "false".
`export CI=true`

GitLab CI Documentation Reference:
- https://docs.gitlab.com/ee/ci/jobs/job_logs.html#custom-collapsible-sections
"""

import logging
import os
import re
import sys
import time
from functools import wraps

from colors import colors


class MaskingFilter(logging.Filter):
    """
    A custom logging filter that dynamically masks sensitive data in log messages.

    This filter allows users to specify sensitive values that will be replaced with
    self.masking_string value whenever they appear in log messages.
    """

    def __init__(self) -> None:
        """
        Initialize the EnhancedMaskingFilter.

        Creates an empty set to store sensitive values and prepares the filter for use.
        """
        super().__init__()
        self.sensitive_values: set[str] = set()  # Store values to mask
        self.masking_string: str = "**[MASKED]**"

    def set_masking_string(self, string: str) -> None:
        """
        Redefine masking string with custom value

        params:
            string: The custom string used to replace masked values.
        """
        if string:
            self.masking_string = string

    def add_sensitive_value(self, value: str) -> None:
        """
        Add a value to the list of sensitive values to be masked.

        params:
            value: The value to be masked in log messages.
        """
        self.sensitive_values.add(re.escape(value))  # Escape special regex characters

    def remove_sensitive_value(self, value: str) -> None:
        """
        Remove a value from the list of sensitive values.

        params:
            value: The value to be removed from masking.
        """
        self.sensitive_values.discard(re.escape(value))

    def mask_value(self, value: str) -> str:
        """
        Mask sensitive values within the given input.

        params:
            value: The value to be checked for sensitive data.
        """
        if not self.sensitive_values:  # Skip masking if no sensitive values are set
            return value

        pattern = re.compile("|".join(self.sensitive_values))
        return pattern.sub(self.masking_string, value)

    def filter(self, record: logging.LogRecord) -> bool:
        """
        Process and mask sensitive data in log records.
        If no sensitive values are defined, the log record is left unchanged.

        params:
            record: The log record to be processed.
        """
        if not self.sensitive_values:  # Skip processing if no sensitive values are set
            return True

        # Handle deferred string formatting
        if record.args:
            # Use a temporary formatter to resolve args into a string
            temp_formatter = logging.Formatter("%(message)s")
            record.msg = temp_formatter.format(record)
            record.args = None  # Prevent reformatting

        if record.msg:
            if not isinstance(record.msg, str):
                temp_formatter = logging.Formatter("%(message)s")
                record.msg = temp_formatter.format(record)

            record.msg = self.mask_value(record.msg)

        return True


def init_masking() -> None:
    """
    Init masking filter
    """

    # MaskingFilter instance
    logging.masking_filter = MaskingFilter()
    # Add the MaskingFilter to the root logger
    logging.getLogger().addFilter(logging.masking_filter)

    logging.set_masking_string = logging.masking_filter.set_masking_string
    logging.add_masked_value = logging.masking_filter.add_sensitive_value
    logging.remove_masked_value = logging.masking_filter.remove_sensitive_value


def init_gitlab_sections(section_header_color: colors = colors.fg.cyan) -> None:
    """
    GitLab supports custom collapsible sections in job logs.
    Reference: https://docs.gitlab.com/ee/ci/jobs/job_logs.html#custom-collapsible-sections
    To use this feature, defined two helper functions and variable.
    If the script is not running in a CI pipeline, sections are printed with delimiters.
    """

    # Detect if the script is running in GitLab CI and attach to the logging module
    logging.GITLAB_CI = os.getenv("CI") == "true"
    logging.section_delimiter = "-" * 80
    logging.section_header_color = section_header_color  # Default header color

    def start_section(
        title: str, header: str, header_color: colors = logging.section_header_color, collapsed: bool = True
    ) -> None:
        """
        Marks the start of a logging section.

        params:
            title: A unique identifier for the section. Only letters, numbers, and "_", "." or "-" allowed.
            header: The text to display at the start of the section.
            header_color: Color for the section header.
            collapsed: Whether the section should be collapsed by default.
        """

        if logging.GITLAB_CI:
            timestamp = int(time.time())
            collapse_flag = "[collapsed=true]" if collapsed else ""
            print(
                f"\033[0Ksection_start:{timestamp}:{title}{collapse_flag}\r\033[0K{header_color}{header}{colors.reset}"
            )
        else:
            # Print the formatted section with the text centered and delimiters on both sides
            section_header = f"  start {header} section  "
            remaining_length = len(logging.section_delimiter) - len(section_header)
            left_length = remaining_length // 2
            right_length = remaining_length - left_length
            print(
                f"{header_color}{logging.section_delimiter[:left_length]}{section_header}{logging.section_delimiter[:right_length]}{colors.reset}"
            )

    def end_section(title: str, header: str, header_color: colors = logging.section_header_color) -> None:
        """
        Marks the end of a logging section.

        params:
            title: A unique identifier for the section.
            header: The text to display at the end of the section.
            header_color: Color for the section header.
        """

        if logging.GITLAB_CI:
            timestamp = int(time.time())
            print(f"\033[0Ksection_end:{timestamp}:{title}\r\033[0K")
        else:
            # Print the formatted section with the text centered and delimiters on both sides
            section_header = f"  end {header} section  "
            remaining_length = len(logging.section_delimiter) - len(section_header)
            left_length = remaining_length // 2
            right_length = remaining_length - left_length
            print(
                f"{header_color}{logging.section_delimiter[:left_length]}{section_header}{logging.section_delimiter[:right_length]}{colors.reset}"
            )

    # Attach functions to the logging module for more convenient usage
    logging.start_section = start_section
    logging.end_section = end_section

    def wrap_log_function(original_log_function):
        """
        Wraps a logging function to optionally include GitLab CI collapsible sections.
        If a section name is provided, the log message will be enclosed
        within a collapsible section in GitLab CI logs or a text-based delimiter otherwise.
        """

        @wraps(original_log_function)
        def wrapper(
            msg,
            *args,
            section_title: str = None,
            section_header: str = None,
            header_color: colors = logging.section_header_color,
            collapsed: bool = True,
            **kwargs,
        ):
            if section_title:
                logging.start_section(section_title, section_header, header_color, collapsed=collapsed)
                original_log_function(msg, *args, **kwargs)
                logging.end_section(section_title, section_header, header_color)
            else:
                original_log_function(msg, *args, **kwargs)

        return wrapper

    # Override logging functions to support GitLab CI sections
    logging.info = wrap_log_function(logging.info)
    logging.debug = wrap_log_function(logging.debug)
    logging.warning = wrap_log_function(logging.warning)
    logging.error = wrap_log_function(logging.error)
    logging.critical = wrap_log_function(logging.critical)


def init_logging(verbose: int, section_header_color: colors = colors.fg.cyan, message_color: colors = "") -> None:
    """
    Configure logging verbosity.

    If running inside GitLab CI, logs can include collapsible sections
    If not separate such sections with text delimiter.

    params:
      verbose: verbosity level.
    """

    base_format = "%(message)s"
    verbosed_format = "%(asctime)s %(name)-30s %(levelname)-8s %(message)s"

    if message_color:
        base_format = f"{message_color}{base_format}{colors.reset}"
        verbosed_format = f"{message_color}{verbosed_format}{colors.reset}"

    if verbose:
        logging.basicConfig(
            level=logging.DEBUG,
            format=verbosed_format,
            datefmt="%H:%M:%S",
            stream=sys.stdout,  # Redirects all logs (INFO, ERROR, etc.) to stdout
            force=True,  # Overwrites previous logging configurations
        )
    else:
        logging.basicConfig(
            level=logging.INFO,
            format=base_format,
            datefmt="%H:%M:%S",
            stream=sys.stdout,  # Redirects all logs (INFO, ERROR, etc.) to stdout
            force=True,  # Overwrites previous logging configurations
        )

    # Initialize useful features
    init_masking()
    init_gitlab_sections(section_header_color)


def main() -> None:
    """
    Example of usage
    """

    def log_something() -> None:
        # Automatically create section
        files = ["file1.txt", "file2.txt", "file3.txt"]
        logging.info(
            "\n".join(files),
            section_title="files_list",
            section_header="Listing files",
            header_color=colors.fg.lightblue,
        )

        # Start/end sections manually
        logging.start_section(
            title="install_dependencies", header="Installing Dependencies", header_color=colors.fg.cyan
        )
        logging.info("Fetching packages...")
        logging.end_section(title="install_dependencies", header="Dependencies Installed", header_color=colors.fg.cyan)

        my_pass = "MySensitiveValue"
        logging.add_masked_value(my_pass)
        logging.info(f"This is my password: {my_pass}")

    # Initialize logging for GitLab CI/CD
    init_logging(verbose=False, message_color=colors.fg.yellow)
    log_something()
    init_logging(verbose=True, message_color=f"{colors.bg.blue}")
    log_something()


if __name__ == "__main__":
    main()
