#!/usr/bin/env python3
import json
import re
import requests
import argparse
from atlassian import Confluence
from html.parser import HTMLParser
from html import unescape
from lxml import html
from lxml.etree import tostring, SubElement
import os
import datetime
import time

class PDFdownloader():
	def __init__(self,token_pdf:str,pdf_base_url:str,file_name:str,job_id_key_pdf:str):
		self.token=token_pdf
		self.base_url=pdf_base_url
		self.job_id_key=job_id_key_pdf
		# Directory where the script is located
		script_dir = os.path.dirname(os.path.abspath(__file__))
		# Relative path to ../data
		data_dir = os.path.join(script_dir, "../data")
		# JSON file name
		json_file_path = os.path.join(data_dir, file_name)
		# Read the JSON data from the file
		try:
			with open(json_file_path, "r", encoding="utf-8") as f:
				# Load JSON into a Python dictionary
				self.request_json = json.load(f)
		except FileNotFoundError:
			print(f"File not found: {json_file_path}")
               
	def get_export_job_id(self,page_id:str)->str:
		#set the pageId to the requested pageid
		self.request_json["pageId"] = page_id
		resp=requests.post( url=f"{self.base_url}/api/public/1/exports" ,headers={"Authorization" : self.token}, json=self.request_json,verify=False)
		resp.raise_for_status()
		resp_dict=resp.json()
		if self.job_id_key in resp_dict:
			self.job_id=resp_dict[self.job_id_key]
			return self.job_id
          
	def get_job_status(self)->dict:
		resp=requests.get( url=f"{self.base_url}/api/public/1/exports/{self.job_id}/status" ,headers={"Authorization" : self.token}, json=self.request_json,verify=False)
		resp.raise_for_status()
		resp_dict=resp.json()
		return resp_dict
    
	def wait_till_url(self,timeout=600)->str:
		start=time.time()
		while not self.get_job_status()["downloadUrl"]:
			print(f"job status: {self.get_job_status()}")
			if time.time() - start > timeout:
				print(f"Timeout to wait for pdf: {timeout}s exceeded")
				return None
			time.sleep(10)	
		return self.get_job_status()["downloadUrl"]
    
	def download_pdf_to_file(self,pg_id:str,filename:str):
		self.get_export_job_id(pg_id)
		download_url=self.wait_till_url()
		response = requests.get(download_url,verify=False)
		response.raise_for_status()
		with open(filename, 'wb') as file:
			file.write(response.content)
		print(f"File from url: {download_url} downloaded into {filename}")

class CustomHelpFormatter(argparse.HelpFormatter):
    def _get_help_string(self, action):
        help = action.help
        if action.default is not argparse.SUPPRESS:
            help += f' (default: {action.default})'
        return help

class TableParser(HTMLParser):
    def __init__(self):
        super().__init__()
        self.tables = []
        self.current_table = None
        self.current_row = None
        self.current_cell = None
        self.in_table = False
        self.in_row = False
        self.in_cell = False

    def handle_starttag(self, tag, attrs):
        if tag == 'table':
            self.current_table = []
            self.tables.append(self.current_table)
            self.in_table = True
        elif tag == 'tr' and self.in_table:
            self.current_row = []
            self.current_table.append(self.current_row)
            self.in_row = True
        elif tag in ('td', 'th') and self.in_row:
            self.current_cell = []
            self.current_row.append(self.current_cell)
            self.in_cell = True

    def handle_endtag(self, tag):
        if tag == 'table' and self.in_table:
            self.in_table = False
        elif tag == 'tr' and self.in_row:
            self.in_row = False
        elif tag in ('td', 'th') and self.in_cell:
            self.in_cell = False

    def handle_data(self, data):
        if self.in_cell:
            self.current_cell.append(data.strip())

def parse_html(html):
    parser = TableParser()
    parser.feed(html)
    return parser.tables

def calculate_test_statistics(tables):
	"""Calculate statistics for test results."""
	test_dict = {}
	results = {"passed": 0, "failed": 0, "not_tested": 0}
	# Skip the first table
	for table in tables[1:]:
		# Skip the header row
		for row in table[1:]:
			sel_col=None
			for col in range(0,2):
				if len(row[col]) > 0 and row[col][0] and isinstance(row[col][0],str) and len(row[col][0].split('.')) >= 3 and any(tst_type in row[col][0] for tst_type in ['PROBE','FUNCTIONAL','STRESS']) and len(row[col][0]) <= 70 and len(row[col][0]) >= 10:
					sel_col=col
					break
			if not sel_col == None:
				if len(row[sel_col+2]) > 0 and isinstance(row[sel_col+2][0],str) and len(row[sel_col+2][0]) > 1 and len(row[sel_col+2][0]) <= 10 and row[sel_col+2][0].lower() in results:
					if not row[col][0] in test_dict: results[row[sel_col+2][0].lower()] += 1 
					result=row[sel_col+2][0].lower()
				else:
					if not row[col][0] in test_dict: results["not_tested"] += 1
					result="not_tested"
				test_dict.update({row[sel_col][0]:result})
	total_tests = len(test_dict.keys())
	percentage = {key: (value / total_tests * 100 if total_tests else 0) for key, value in results.items()}
	return results, percentage, test_dict

def generate_summary_table(results, percentage, unique_tests):
    """Generate an HTML table for the summary of test results."""
    summary_table = f'''<div><h3>Summary Table</h3>

    <table data-table-width="1680" data-layout="default">
    <colgroup><col style="width: 300.0px;"/><col style="width: 120.0px;"/><col style="width: 120.0px;"/></colgroup>
    <tbody>
    <tr><th>Test Result</th><th>Count</th><th>Percentage</th></tr>
    <tr><td>Passed</td><td>{results["passed"]}</td><td>{percentage["passed"]:.2f}%</td></tr>
    <tr><td>Failed</td><td>{results["failed"]}</td><td>{percentage["failed"]:.2f}%</td></tr>
    <tr><td>Not Tested</td><td>{results["not_tested"]}</td><td>{percentage["not_tested"]:.2f}%</td></tr>
    </tbody>
    </table    
    <h3>Tests</h3>
    <table data-table-width="1680" data-layout="default">
    <colgroup><col style="width: 300.0px;"/><col style="width: 120.0px;"/></colgroup>
    <tbody>
    <tr><th>Test Name</th><th>Result</th></tr>
    '''
    
    for test in unique_tests:
        # Default to "not tested" unless overridden
        result = "not_tested"
        if "failed" in unique_tests[test].lower():
            result = "failed"
        elif "passed" in unique_tests[test].lower():
            result = "passed"

        if result == "failed":
            result_html = '''<div style="border: 2px solid red; background-color: #ffcccc; color: black; padding: 5px; text-align: center;">Failed</div>'''
        elif result == "passed":
            result_html = '''<div style="border: 2px solid green; background-color: #ccffcc; color: black; padding: 5px; text-align: center;">Passed</div>'''
        else:
            result_html = '''<div style="border: 2px solid yellow; background-color: #ffffcc; color: black; padding: 5px; text-align: center;">Not Tested</div>'''

        summary_table += f"<tr><td>{test}</td><td>{result_html}</td></tr>"

    summary_table += "</tbody></table></div>"
    return summary_table

def insert_summary_table(page_content, summary_html):
    """Insert the summary table before the next table's title."""
    tree = html.fromstring(page_content)
    body = tree.xpath("//body")[0]
    all_tables = body.xpath("//table")
    next_title = all_tables[1].getprevious().getprevious() if len(all_tables) > 1 else None
    summary_element = html.fragment_fromstring(summary_html)
    # Assuming titles are h3 tags
    if next_title is not None and next_title.tag == 'h3':
        next_title.addprevious(summary_element)
    elif all_tables:
        all_tables[1].addprevious(summary_element)
    else:
        body.insert(0, summary_element)
    return tostring(tree, encoding='unicode', method='xml')

def get_table_content(tables, table_number, row_number, col_number):
    try:
        return tables[table_number][row_number][col_number]
    except IndexError:
        return None

def tokenize(text):
    # Convert the text to lowercase and split into words
    return set(text.lower().split())

def most_common_words(target, candidates):
    target_words = tokenize(target)
    max_common_count = 0
    best_match = None
    
    for candidate in candidates:
        candidate_words = tokenize(candidate)
        common_words = target_words.intersection(candidate_words)
        common_count = len(common_words)
        
        if common_count > max_common_count:
            max_common_count = common_count
            best_match = candidate
    
    return best_match

def get_serial_and_rev(host, jobid):
	# API endpoints
	job_url = f"{host}/api/v0.2/jobs/{jobid}"
	# Perform API call to get actual_device from the job_url
	response_job = requests.get(job_url,timeout=10)
	job_data = response_job.json()
	actual_device = job_data.get("actual_device")
	# Perform API call to get tags from the actual_device_url
	if actual_device:
		device_url_template = f"{host}/scheduler/device/{actual_device}"
		try:
			# Make an HTTP GET request
			response = requests.get(device_url_template)
			response.raise_for_status()  # Raise an error for HTTP requests with status codes 4xx or 5xx
			# Extract the "Tags" section using regex
			tags_match = re.search(r'<dt>Tags</dt>\s*<dd>(.*?)</dd>', response.text, re.DOTALL)
			if not tags_match:
				return ""
			# Extract individual tags and clean them up
			tags_content = tags_match.group(1)
			tags = re.findall(r'[\w\-]+', tags_content)
			final_t=[str(t) for t in tags if str(t).startswith(("K","rev","REV","Rev"))]
			return ','.join(final_t)
		except requests.exceptions.RequestException as e:
			return f"Error fetching the page: {e}"

def sanitize_text(text):
    """Remove control characters and invalid XML characters from a string."""
    if not text:
        return ""
    # Remove control characters except \t, \n, \r
    text = re.sub(r'[^\x09\x0A\x0D\x20-\x7E\u0080-\uFFFF]', '', text)
    # Ensure special XML characters are escaped
    text = text.replace('&', '&amp;').replace('<', '&lt;').replace('>', '&gt;')
    return text

def update_table_content(page_content:str, table_number:int, row_number:int, col_number:int, new_content:str, res_final=None)->str:
    tree = html.fromstring(page_content)
    tables = tree.xpath('//table')
    if table_number >= len(tables):
        raise Exception(f"Table number {table_number} not found")
    table = tables[table_number]
    rows = table.xpath('.//tr')
    if row_number >= len(rows):
        raise Exception(f"Row number {row_number} not found in table {table_number}")

    cells = rows[row_number].xpath('.//td')
    if col_number >= len(cells):
        raise Exception(f"Column number {col_number} not found in row {row_number} of table {table_number}")
    
    cell = cells[col_number]
    cell.clear()  # Clear existing content

    if res_final:
        # Add a rectangular paragraph with rounded borders and conditional formatting
        border_color = "green" if res_final == "passed" else "red"
        filler_color = "lightgreen" if res_final == "passed" else "pink"
        content = f"""
        <div style="
            border: 2px solid {border_color};
            border-radius: 8px;
            background-color: {filler_color};
            padding: 10px;
            text-align: center;
            color: {border_color};
            font-weight: bold;">
            {res_final.capitalize()}
        </div>
        """
        cell.append(html.fragment_fromstring(content, create_parent='div'))
    elif new_content:
        # Default behavior: handle new_content with plain text or multiple paragraphs
        if '\n' in new_content or ('<' in new_content and '>' in new_content):
            lines = new_content.split('\n')
            for line in lines:
                sanitized_line = sanitize_text(line)  # Remove invalid characters
                if '<' in sanitized_line and '>' in sanitized_line:
                    parsed_html = html.fragment_fromstring(sanitized_line, create_parent='div')
                    cell.append(parsed_html)
                else:
                    p = SubElement(cell, 'p')
                    p.text = sanitized_line
        else:
            cell.text = sanitize_text(new_content)  # Sanitize before assignment
    else:
        # Handle None or empty new_content gracefully
        cell.text = ""

    return tostring(tree, encoding='unicode', method='xml')


def push_page(confl:Confluence,val_rep_space,templ_title,content:str,pdf_downdr:PDFdownloader):
	r_list=["Validation Report Template", "Clea OS" ,"Embedded", "Things"]
	parent_title=re.sub(r'\b(?:' + '|'.join(map(re.escape, r_list)) + r')\b', '', templ_title).strip() + " Weekly Reports"
	parent_id=confl.get_page_id(val_rep_space,parent_title)
	new_title=templ_title.replace("Template","") + "Automation " + datetime.datetime.now().strftime("%d-%m-%Y")
	confluence.update_or_create(parent_id,new_title,content,editor="v2",representation="storage")
	new_page_id=confluence.get_page_id(val_rep_space, new_title)
	time.sleep(2)
	new_url=f"<h3 style='text-decoration: none;color: #0078D7;font-weight: bold;'> Confluence Report Link: {confl.url}/spaces/{val_rep_space}/pages/{new_page_id} </h3>"
	content=new_url + content
	confluence.update_or_create(parent_id,new_title,content,editor="v2",representation="storage")
	pdf_downdr.download_pdf_to_file(new_page_id,'report.pdf')
	return f"{confl.url}/spaces/{val_rep_space}/pages/{new_page_id}"


def get_pages_with_regex(confl:Confluence,format:re)->list:
	pages_titles=[]
	count=0
	while True:
		count=count+1
		pages=confl.get_space_content(args.validation_page_space,limit=100,start=(count*100))
		results=pages['page']['results']
		if not results:
			break
		pages_titles=pages_titles + [ res["title"] for res in results if "title" in res]
	single_line_titles="\n".join(entry for entry in pages_titles)
	matches = re.findall(format, single_line_titles, re.MULTILINE)
	return matches
    

def clean_json_line(line: str) -> str:
    """
    Clean up a JSON line by handling invalid escape sequences, removing binary markers,
    and fixing other formatting issues.
    """
    # Remove invalid escape sequences
    line = re.sub(r'\\x[0-9A-Fa-f]{2}', '', line)  # Remove \xHH
    line = re.sub(r'\\\\', '', line)  # Remove double backslashes
    # Clean up invalid characters
    line = re.sub(r'[!]', '', line)  # Remove invalid `!`
    line = re.sub(r'\\0', '', line)  # Remove `\0` escape sequences
    return line

def extract_last_occurrences_optimized(line_dict, regex_dict):
    """
    Optimized function to extract the last occurrence of a match for each test's regex in line_dict.

    Args:
        line_dict (list of dict): List of dictionaries containing the "msg" key.
        regex_dict (dict): Dictionary where keys are test names and values are regex patterns.

    Returns:
        dict: Dictionary with test names as keys and the last matched value as values.
    """
    # Pre-compile all regex patterns for efficiency
    compiled_regex = {
        test_name: [re.compile(pattern, re.DOTALL | re.MULTILINE) for pattern in patterns]
        for test_name, patterns in regex_dict.items()
    }

    # Initialize a dictionary to store the last matches
    last_matches = {test_name: None for test_name in regex_dict}

    # Iterate over the messages in reverse order to find the last occurrence directly
    for entry in reversed(line_dict):
        msg = entry.get("msg", "")
        if not isinstance(msg, str):
            continue

        for test_name, patterns in compiled_regex.items():
            for pattern in patterns:
                match = pattern.search(msg)
                if match:
                    # Save the match and stop processing further for this test name
                    if isinstance(match.group(), str):
                        last_matches[test_name] = match.group()
                    elif isinstance(match.groups(), tuple):
                        last_matches[test_name] = "".join(match.groups())
                    break

    # Filter out None values for tests with no matches
    return {key: value for key, value in last_matches.items() if value is not None}




def clean_msg_list(msg_dict: dict):
    for msg_key in list(msg_dict.keys()):
        try:
            # Try to find the index of the first element containing `msg_key`
            start_index = next(i for i, s in enumerate(msg_dict[msg_key]) if msg_key in s)
            # Update the list to start from the found index
            msg_dict[msg_key] = msg_dict[msg_key][start_index:]
        except StopIteration:
            # If no match is found, you can either skip this key or assign an empty list
            msg_dict[msg_key] = []  # or use `continue` to skip
    return msg_dict




def parse_log_file(file_path: str) -> list:
    parsed_logs = []
    try:
        with open(file_path, 'r') as file:
            for line in file:
                try:
                    # Parse the cleaned JSON string
                    log_entry = clean_json_line(line.strip()[2:])
                    parsed_logs.append(json.loads(log_entry))
                except json.JSONDecodeError as e:
                    print(f"Error decoding JSON: {e} - Line: {line.strip()[2:]}")
    except FileNotFoundError as e:
        print(f"File not found: {e}")
    return parsed_logs

def download_file(pre_job_url:str,job_ID:str,post_job_url:str,path:str)->str:
    url=f"{pre_job_url}{job_ID}{post_job_url}"
    r = requests.get(url) 
    try:
        with open(f"{path}log-{job_ID}.txt",'wb') as f: 
            f.write(r.content) 
        return f"{path}log-{job_ID}.txt"
    except Exception as e:
        print(e)
        return None

def get_job_id_from_artifacts():
    """
    Recursively searches in the folder external to the script's directory and all its subdirectories
    for a file with the format: results-job_id-job_platform.xml.

    Returns:
        str: The job ID extracted from the filename if the file is found.

    Raises:
        FileNotFoundError: If no matching file is found.
        ValueError: If a file is found but its format is invalid.
    """
    # Calculate the external folder path
    script_dir = os.path.dirname(os.path.abspath(__file__))
    folder_to_search = os.path.abspath(os.path.join(script_dir, "../../"))
    # Define the regex pattern for the file format
    format = r"results-(\d+)-.*\.xml"
    pattern = re.compile(format)

    # Check if the folder exists
    if not os.path.isdir(folder_to_search):
        raise FileNotFoundError(f"The folder '{folder_to_search}' does not exist.")
    # Walk through the folder and its subdirectories
    for root, _, files in os.walk(folder_to_search):
        for filename in files:
            match = pattern.fullmatch(filename)
            if match:
                # Extract and return the job ID as a string
                return str(match.group(1))
    # Raise an error if no valid file is found
    raise FileNotFoundError(
        f"No file with the required format: {format} was found in the folder: {folder_to_search} or its subdirectories."
    )

def parse_args(deflt_args:dict):
	parser = argparse.ArgumentParser(description="Parse LAVA Logs",formatter_class=CustomHelpFormatter)
	parser.add_argument("--board_name",required=True,type=str, help="put the board name to search inside the confluence pages")
	parser.add_argument("--distro",type=str,required=True, help="the distro to search the validation report template from")
	parser.add_argument("--pre_job_url", type=str,default="http://lavalab.seco.com/scheduler/job/", help="put the url to download the job plaintext file, until the job id part")
	parser.add_argument("--post_job_url", type=str,default="/log_file/plain",help="put the end part of the url to download the job plaintext file, from the job-id onward")
	parser.add_argument("--confluence_url", type=str,default="https://secogroup.atlassian.net",help="confluence base url")
	parser.add_argument("--username",type=str,required=True, help="username for the confluence page")
	parser.add_argument("--token",type=str,required=True, help="token for the confluence page")
	parser.add_argument("--validation_page_space",type=str,default="SW",help="validation page space")
	parser.add_argument("--pdf_base_url",type=str,default="https://scroll-pdf.us.exporter.k15t.app", help="template page title")
	parser.add_argument("--pdf_token",type=str,required=True, help="token to authenticate to the K15T pdf API")
	parser.add_argument("--pdf_json_name", type=str, default="pdf_downloader.json", help="JSON file name from which the program can take the body of the K15T export job initializzation request")
	parser.add_argument("--pdf_job_id_key", type=str, default="jobId", help="the job id key in the JSON that the K15T API responds back upon correct export job initializzation")
	# Getting first key in dictionary
	arg1 = next(iter(deflt_args))
	#Getting the first value in the dictionary
	def1 = deflt_args[arg1]
	parser.add_argument("--"+str(arg1), type=str, default=str(def1),help="put the path to the folder in which the log will be downlaoded")
	return parser.parse_args()

def Get_RM_Dicts(tst_l: list) -> dict:
    r_dict = {}
    m_dict = {}
    msg_list=[]
    for test_l_dict in tst_l:
        if not test_l_dict:  # Check if test_l_dict is empty
            continue  # Skip empty lists
        if isinstance(test_l_dict["msg"],str):
            msg_list.append(test_l_dict["msg"])
        elif test_l_dict["lvl"] == "results":
            m_dict.update({test_l_dict["msg"]["case"]:msg_list})
            r_dict.update({test_l_dict["msg"]["case"]:test_l_dict["msg"]["result"]})
            msg_list=[]
    return r_dict, m_dict

def print_results(res_dict:dict,msg_dict:dict)->None:
    for test_name in list(res_dict.keys()):
        result=res_dict[test_name]
        msg_list=msg_dict[test_name]
        print("\n\n////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////"+str(test_name)+" : "+str(result)+"////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////\n")
        print("\t"+"\n\t".join(msg_list))
    
def rm_file(file_path:str):
    if os.path.isfile(file_path):
        os.remove(file_path)
    else:
        # If it fails, inform the user.
        print("Error: %s file not found" % file_path)

def get_page_tables(confluence:Confluence,space_key:str,title:str):
    page=confluence.get_page_by_title(space_key,title,expand="body.storage")
    html=page["body"]["storage"]["value"]
    html_content = unescape(html)
    # Parse the HTML content
    tables = parse_html(html_content)
    return tables,html

def build_test_dicts(tab:list):
    test_dict={}
    for i in range(len(tab)):
        for j in range(len(tab[i])):
            for k in range(len(tab[i][j])):
                if len(tab[i][j][k])>0:
                    if len(tab[i][j][k][0].split())==1:
                        if len(tab[i][j][k][0].split("."))>1:
                            if tab[i][j][k][0].split(".")[1] in ["STRESS","FUNCTIONAL","PROBE"]:
                                if tab[i][j][k][0] in test_dict and isinstance(test_dict[tab[i][j][k][0]],list):
                                    old_list=test_dict[tab[i][j][k][0]]
                                else:
                                    old_list=[]
                                old_list.append([i,j,k])
                                test_dict.update({tab[i][j][k][0]:old_list})
                                break
                    if i==0 and k == 0:
                        # this is because it's found in the second column whilst the usual results are in the 4th one
                        test_dict.update({tab[i][j][k][0]:[i,j,k]})
    return test_dict

def generate_new_tables(page_cnt,tst_dict:dict,msg_dict:dict,rs_dict:dict,last_occ:dict) -> list:
    lst_matching_tst=[ str(tst_k) for tst_k in list(tst_dict.keys()) if any(tst_k in ms_k for ms_k in list(msg_dict.keys())) ]
    updated_tables=page_cnt
    for match_tst in lst_matching_tst:
        if match_tst in list(last_occ.keys()):
            positions=tst_dict[match_tst]
            if not all(isinstance(pos,list) for pos in positions):
                positions=[positions]
            for pos in positions:
                updated_tables=update_table_content(updated_tables,pos[0],pos[1],pos[2]+1,msg_dict[match_tst])
        else:
            matched_tests=[ mk for mk in list(msg_dict.keys()) if match_tst in mk ]
            log_match=""
            res_final=True
            for mtc_tst in matched_tests:
                log_match += "\n################################\n" + str(mtc_tst) + "\n################################\n" + "\n$ ".join(msg_dict[mtc_tst]) + "<br>"
                res_final=res_final and (rs_dict[mtc_tst] == "pass")
            res_final="passed" if res_final else "failed"
            positions=tst_dict[match_tst]
            if not all(isinstance(pos,list) for pos in positions):
                positions=[positions]
            for pos in positions:
                updated_tables=update_table_content(updated_tables,pos[0],pos[1],pos[2]+3,log_match)
                updated_tables=update_table_content(updated_tables,pos[0],pos[1],pos[2]+2,"",res_final)
    return updated_tables

if __name__ == "__main__":
	
	#############################Defaults#############################
	# Path to the log file, this program will replace #JOB-ID# with the actual job-id requested
	default_args={'dw_dir':'/tmp/'}
	regex_dict = {
		"U-Boot Main Version": [r"U-Boot \d{4}\.\d{2}.*?$"],
		"Kernel Main Version": [r"Linux version \d+\.\d+\.\d+.*?$"],
		"OS Version": [r"Clea OS.*? \d+\.\d+\.\d+"],
		"Filesystem Main Version": [r"seco.*?wic\.bz2"],
		"Software Artifacts": [r"https.*?seco.*?wic\.bz2"],
		"Additional SW Tools": [r"(https:\/\/).*?(git\.seco\.com.*?lava-test\.git).*?( checkout.*?$)"],
	}
	job_id=get_job_id_from_artifacts()
	
	#####################API Classes Configuration########################
	#########Confluence##########
	args=parse_args(default_args)
	confluence = Confluence(
		url=args.confluence_url,
		username=args.username,
		password=args.token,
	)
	confluence.session.verify = False
	#########K15T pdf Downloader##########
	pdf_downloader=PDFdownloader(f"Bearer {args.pdf_token}",args.pdf_base_url,args.pdf_json_name,args.pdf_job_id_key)
	
	####Validation Report page search#####
	matched_titles=get_pages_with_regex(confluence,rf".*{str(args.board_name).upper()}.*Clea OS.*Validation Report Template")
	distro = "Embedded" if "embedded" in args.distro else "Things" if "things" in args.distro else (_ for _ in ()).throw(ValueError("Invalid distro"))
	template_title=[ mtc_titl for mtc_titl in matched_titles if distro in mtc_titl][0]
	
	#######################Post Args Variables########################
	post_args_known={
		"LAVA Test Log":f"{args.pre_job_url}{job_id}",
		"Tester" : "GitLab Automation",
		"Report latest Update" : datetime.datetime.now().strftime("%d-%m-%Y"),
		"Source Code Repository" : f"{os.getenv('CI_SERVER_URL','')}{os.getenv('CI_PROJECT_PATH','')}" if os.getenv('CI_SERVER_URL',None) and os.getenv('CI_PROJECT_PATH',None) else '',
		"Source Code Branch" : os.getenv('CI_COMMIT_BRANCH',''),
		"Source Code Tag/Commit": os.getenv('CI_COMMIT_TAG','') if os.getenv('CI_COMMIT_TAG',None)  else os.getenv('CI_COMMIT_SHORT_SHA',''),
		"GitLab Pipeline" : os.getenv('CI_PIPELINE_URL',''),
		"HW References (PN, SN)" : get_serial_and_rev("/".join(str(args.pre_job_url).split("/")[:-3]),job_id)
	}
	##################################################################
	
	folder_path=getattr(args,next(iter(default_args)))
	dw_path=download_file(args.pre_job_url,job_id,args.post_job_url,folder_path)
	if dw_path:
		parsed_logs=parse_log_file(dw_path)
		last_occurrences = extract_last_occurrences_optimized(parsed_logs, regex_dict)
		last_occurrences = last_occurrences | post_args_known
		result_dict,message_dict=Get_RM_Dicts(parsed_logs)
		message_dict=clean_msg_list(message_dict)
		message_dict=message_dict | last_occurrences
		tables,page=get_page_tables(confluence,args.validation_page_space,template_title)
		test_dict=build_test_dicts(tables)
		new_html=generate_new_tables(page,test_dict,message_dict,result_dict,last_occurrences)
		new_tables=parse_html(new_html)
		results, percentage, unique_tests = calculate_test_statistics(new_tables)
		summary_html = generate_summary_table(results, percentage, unique_tests)
		updated_html = insert_summary_table(new_html, summary_html)
		url=push_page(confluence,args.validation_page_space,template_title,updated_html,pdf_downloader)
		print(f"Confluence Report URL: {url}")
		rm_file(dw_path)
	else:
		print(f"Failed, to download from {args.pre_job_url}{job_id}{args.post_job_url} in {args.next(iter(default_args))}") 